@@ -1,22 +1,22 @@
 import dataclasses
 import logging
-import random
 import re
-import string
 import traceback
 from dataclasses import dataclass
 from datetime import datetime
 from pathlib import Path
 from textwrap import dedent, indent
-from typing import Iterator, Optional, Union
+from typing import Any, Iterator, Union
 
 import yaml
 from yaml import YAMLError
 
+import tgpy.api
 from tgpy import app
-from tgpy.run_code import meval
-from tgpy.run_code.utils import format_traceback
-from tgpy.utils import MODULES_DIR, filename_prefix
+from tgpy._core.utils import format_traceback
+from tgpy.api import MODULES_DIR, STD_MODULES_DIR
+from tgpy.api.tgpy_eval import tgpy_eval
+from tgpy.utils import FILENAME_PREFIX
 
 logger = logging.getLogger(__name__)
 
@@ -39,7 +39,20 @@ def get_module_names() -> Iterator[str]:
         yield file.stem
 
 
-def get_sorted_modules() -> 'list[Module]':
+def get_std_modules() -> 'list[Module]':
+    disabled_modules = tgpy.api.config.get('core.disabled_modules', [])
+    modules = []
+    for file in STD_MODULES_DIR.iterdir():
+        if file.suffix != '.py':
+            continue
+        mod_name = file.stem
+        if mod_name not in disabled_modules:
+            modules.append(Module.load(mod_name, str(file)))
+    modules.sort(key=lambda mod: mod.priority)
+    return modules
+
+
+def get_user_modules() -> 'list[Module]':
     modules = []
 
     for mod_name in get_module_names():
@@ -57,13 +70,15 @@ def get_sorted_modules() -> 'list[Module]':
 
 
 async def run_modules():
-    for module in get_sorted_modules():
+    for module in get_std_modules():
+        await module.run()
+    for module in get_user_modules():
         # noinspection PyBroadException
         try:
             await module.run()
         except Exception:
             logger.error(f'Error during running module {module.name!r}')
-            logger.error(''.join(format_traceback()))
+            logger.error(format_traceback()[1])
             continue
 
 
@@ -82,6 +97,8 @@ def serialize_module(module: Union['Module', dict]) -> str:
     else:
         module_dict = module
     module_code = DOCSTRING_RGX.sub('', module_dict.pop('code')).strip()
+    module_dict.update(module_dict.get('extra', {}))
+    module_dict.pop('extra', None)
     module_str_metadata = yaml.safe_dump(module_dict).strip()
     return MODULE_TEMPLATE.format(
         metadata=indent(module_str_metadata, ' ' * 4),
@@ -89,13 +106,12 @@ def serialize_module(module: Union['Module', dict]) -> str:
     )
 
 
-def deserialize_module(data: str, name: Optional[str] = None) -> 'Module':
-    name = name or f'<unknown module {"".join(random.choices(string.digits, k=8))}>'
+def deserialize_module(data: str, name: str) -> 'Module':
     docstring_match = DOCSTRING_RGX.search(data)
-    fallback_metadata = {
+    fallback_metadata: dict[str, Any] = {
         'name': name,
-        'origin': f'{filename_prefix}module/{name}',
-        'priority': datetime.now().timestamp(),
+        'origin': f'{FILENAME_PREFIX}module/{name}',
+        'priority': int(datetime.now().timestamp()),
     }
     if docstring_match:
         module_str_metadata = dedent(
@@ -113,6 +129,12 @@ def deserialize_module(data: str, name: Optional[str] = None) -> 'Module':
         logger.warning(f'No metadata found in module {name!r}')
     # to support debugging properly, module code is the whole file
     module_dict['code'] = data
+    field_names = {x.name for x in dataclasses.fields(Module)}
+    extra_fields = set(module_dict.keys()) - field_names
+    if extra_fields:
+        module_dict['extra'] = module_dict.get('extra', {})
+        for field_name in extra_fields:
+            module_dict['extra'][field_name] = module_dict.pop(field_name)
     module = Module(**module_dict)
     return module
 
@@ -124,12 +146,13 @@ class Module:
     origin: str
     priority: int
     once: bool = False
-    save_locals: bool = True
+    extra: dict = dataclasses.field(default_factory=dict)
 
     @classmethod
-    def load(cls, mod_name: str) -> 'Module':
-        filename = get_module_filename(mod_name)
-        with open(filename, 'r') as f:
+    def load(cls, mod_name: str, filename: str | None = None) -> 'Module':
+        if not filename:
+            filename = get_module_filename(mod_name)
+        with open(filename, 'r', encoding='utf-8') as f:
             module = deserialize_module(f.read(), mod_name)
         if module.name != mod_name:
             raise ValueError(
@@ -140,20 +163,16 @@ class Module:
     def save(self):
         filename = get_module_filename(self.name)
 
-        with open(filename, 'w') as f:
+        with open(filename, 'w', encoding='utf-8') as f:
             f.write(serialize_module(self))
 
     async def run(self):
-        new_variables, result = await meval(
+        # noinspection PyProtectedMember
+        app.ctx._set_is_module(True)
+        await tgpy_eval(
             self.code,
-            self.origin,
-            globals(),
-            app.api.variables,
-            msg=None,
-            print=lambda *args, **kwargs: None,
-            **app.api.constants,
+            message=None,
+            filename=self.origin,
         )
-        if self.save_locals:
-            app.api.variables.update(new_variables)
         if self.once:
             delete_module_file(self.name)
